#!/usr/bin/perl -w
use strict;
use Math::Trig;
use OpenGL qw/ :all /;
use OpenGL::Image;

eval 'use Time::HiRes qw( gettimeofday )';
my $hasHires = !$@;

$|++;


# Define constants
use constant DEBUG=>0;

# Key map
my $keys =
{
  '!'=>'1',
  '@'=>'2',
  '#'=>'3',
  '$'=>'4',
  '%'=>'5',
  '^'=>'6',
  '&'=>'7',
};

# Get image file
my($path,@opts) = @ARGV;
die qq
{
USAGE: $0 [IMAGE_FILE | PATH] [OPTIONS]

IMAGE_FILE	Identifies a single source image for tiling
PATH		Recursed path for randomly selected source images

OPTIONS:
w=WIDTH     Width of display window     - If neither w nor h are present, a size
h=HEIGHT    Height of display window      is calculated from the source image.

} if (!$path);
die "$path does not exist\n" if (!-e $path);


# Get options
my $opts = {};
foreach my $opt (@opts)
{
  my($key,$value) = split('=',$opt);
  next if (!$key);
  $opts->{$key} = defined($value) ? $value : 0;
}

# Default options
my $sort = $opts->{sort} || 'random';
my $dur = defined($opts->{d}) ? $opts->{d} : 60;
my $fps = defined($opts->{fps}) ? $opts->{fps} : 30;
my $fs = 0;
my $scl = 1.0;
my $sgn = 1.0;
my $display_source = 0;
my $freqx = 1;
my $freqy = 1;
my $equilateral = 0;
my $pad = 5;

if (0)
{
# Get image path(s)
my @images = GetImages($path);
my $images = scalar(@images);
die "No images (jpg,png,gif,tga,bmp) found\n" if (!$images);
my $current = 0;
}

# Get app name
$0 =~ m|^([^\.]+)|;
my $name = $1 || 'capture';

# Window parameters
my $wnd_ID;
$0 =~ m|([^\.]+)|;
my $wnd_title = lc($1);
my($save_w,$save_h,$save_x,$save_y);

# State parameters
my $last_time = $hasHires ? gettimeofday() : 0;
my $last_image = 0;
my $frames = 0;
my $pause = 1;
my($image,$image_w,$image_h);
my($x0,$y0,$x1,$y1,$x2,$y2);
my($ix,$iy,$ix0,$iy0,$ix1,$iy1,$ix2,$iy2);
my($dx,$dy,$dx0,$dy0,$dx1,$dy1,$dx2,$dy2);

# Init GLUT
glutInit();
glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH);
glutInitWindowSize(1,1);

# Open a window
$wnd_ID = glutCreateWindow($wnd_title);
#glutHideWindow();

# Test for necessary OpenGL Extensions
# Must do this _after_ window context is established
my $stat = OpenGL::glpCheckExtension('GL_ARB_texture_rectangle');
die "GL_ARB_texture_rectangle not supported on this video card\n" if ($stat);
$stat = OpenGL::glpCheckExtension('GL_EXT_framebuffer_object');
die "GL_EXT_framebuffer_object not supported on this video card\n" if ($stat);

# Alloc textures
my($tex_ID,$fbo_ID) = glGenTextures_p(2);
Terminate("Unable to alloc texture IDs") if (!$tex_ID || !$fbo_ID);
my $tex_mode = GL_TEXTURE_RECTANGLE_ARB;
glBindTexture($tex_mode, $tex_ID);

# Load image
my($img_type,$img_format,$img_size);
LoadImage($path);
die "Unable to load $path\n" if (!$image);

# Initiate frame buffer
my $fbo_max;
eval('$fbo_max = glGetIntegerv_p(GL_MAX_RENDERBUFFER_SIZE_EXT)');
$fbo_max = 4096 if (!$fbo_max);
my $fbo_w = $image_w * 3;
my $fbo_h = $image_h * 2;

if ($fbo_w > $fbo_max)
{
  $fbo_h = int($fbo_h * $fbo_max / $fbo_w);
  $fbo_w = $fbo_max;
}

my($fb_ID) = glGenFramebuffersEXT_p(1);
my($rb_ID) = glGenRenderbuffersEXT_p(1);
InitFrameBuffer();

# Tiling modes
my $modes = {'0'=>1,'1'=>2,'2'=>1};
my $mode = 0;
my $mfreqx = 1;
my $mfreqy = 1;
my $mx = $fbo_w / 4;
my $my = $fbo_h / 4;
my($ml) = ($mx <= $my) ? $mx : $my;
my $ma = rand(1);
#my $mda = rand(1)/200;
my $mda = .0005;

# Set default Window size
if (!$opts->{w} && !$opts->{h})
{
  $opts->{w} = 512;
  $opts->{h} = 128 * 3;
}
elsif (!$opts->{h})
{
  $opts->{h} = $opts->{w} * 3 / 2;
}
elsif (!$opts->{w})
{
  $opts->{w} = int(.5 + $opts->{h} * 2 / 3);
}

# Resize window
my $wnd_width = $opts->{w};
my $wnd_height = $opts->{h};
print "Image size: $image_w x $image_h\n";
print "FBO size: $fbo_w x $fbo_h\n";
print "Window size: $wnd_width x $wnd_height\n";
glutReshapeWindow($wnd_width,$wnd_height);
#glutShowWindow();

# Register rendering callback
glutDisplayFunc(\&cbRenderScene);

# Register idle callback
glutIdleFunc(\&cbRenderScene);

# Register resize callback
glutReshapeFunc(\&cbResizeScene);

# Register keyboard callback
glutKeyboardFunc(\&cbKeyPressed);

# Init app environment
InitViewport($wnd_width,$wnd_height);

# Print keyboard commands
print qq
{
'p' to toggle play/pause; 'r' to rewind animation;
'+' for faster; '-' for slower; '*' to unthrottle framerate;
'1'-'9' sets tiling frequency; 's' to save as JPEG;
'm' to step through tiling modes; 'q' or <esc> to quit.

OpenGL window must have focus for input.

};

# Pass off control to GLUT
glutMainLoop();
exit(0);


########
# Subroutines
########

# Cleanup routine
sub Terminate
{
  # Disable app
  glutHideWindow();
  glutKeyboardFunc();
  glutSpecialFunc();
  glutIdleFunc();
  glutReshapeFunc();

  glDeleteTextures_p($tex_ID) if ($tex_ID);

  # Now we can destroy window
  glutDestroyWindow($wnd_ID);
}

# Load image
sub LoadImage
{
  my($path) = @_;

  $image = new OpenGL::Image(engine=>'Magick',source=>$path);
  return if (!$image);

  ($image_w,$image_h) = $image->Get('width','height');
  return if (!$image_w || !$image_h);

  # Init texture coords
  InitCoords();

  # Init texture velocity
  $dx = $image_w/100;
  $dy = $image_h/100;
  $ix0 = rand($dx)-$dx;
  $iy0 = rand($dy)-$dy;
  $ix1 = rand($dx)-$dx;
  $iy1 = rand($dy)-$dy;
  $ix2 = rand($dx)-$dx;
  $iy2 = rand($dy)-$dy;

  # Scale velocity
  $dx0 = $scl * $ix0;
  $dy0 = $scl * $iy0;
  $dx1 = $scl * $ix1;
  $dy1 = $scl * $iy1;
  $dx2 = $scl * $ix2;
  $dy2 = $scl * $iy2;

  ($img_type,$img_format,$img_size) =
    $image->Get('gl_internalformat','gl_format','gl_type');

  glTexImage2D_c($tex_mode, 0, $img_type, $image_w, $image_h,
    0, $img_format, $img_size, $image->Ptr());

  glBindTexture($tex_mode, $tex_ID);
  glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_DECAL);
}

# Init sample coords
sub InitCoords
{
  # Note: padding to workaround edge-clamping issues on some OpenGL drivers
  $x0 = $pad;
  $y0 = $pad;
  $x1 = $image_w / 2.0;
  $y1 = $equilateral ? ($x1-($pad*2)) * tan(3.1415/3) : ($image_h - $pad);
  $x2 = $x1 * 2.0 - $pad;
  $y2 = $pad;
}

# Initialize FrameBuffer
sub InitFrameBuffer
{
  glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, $fb_ID);
  glBindTexture($tex_mode, $fbo_ID);

  glTexImage2D_c($tex_mode, 0, $img_type, $fbo_w, $fbo_h,
    0, $img_format, $img_size, 0);

  glTexParameteri($tex_mode, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glTexParameteri($tex_mode, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

  # Bind texture/frame/render buffers
  glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT,
    $tex_mode, $fbo_ID, 0);

# Necessary?
if (0)
{
  glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, $rb_ID);
  glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT, GL_DEPTH_COMPONENT24,
    $fbo_w, $fbo_h);
  glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT,
    GL_RENDERBUFFER_EXT, $rb_ID);
}

  #$stat = glCheckFramebufferStatusEXT(GL_RENDERBUFFER_EXT);
  #printf("FBO Status: %04X\n",$stat);
}

# Rendering callback
sub cbRenderScene
{
  # Throttle FPS
  if ($hasHires && ($fps > 0))
  {
    my $spf = 1 / $fps;
    return if ($spf > (gettimeofday() - $last_time));
    $last_time = gettimeofday();
  }

  # Set up rendering
  glLoadIdentity();
  glBindTexture($tex_mode, $tex_ID);

  glTexParameteri($tex_mode, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
  glTexParameteri($tex_mode, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
  glTexParameteri($tex_mode, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glTexParameteri($tex_mode, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

  glEnable($tex_mode);

  # Display source image
  if ($display_source)
  {
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);

    glBegin (GL_QUADS);
    {
      glTexCoord2f(0,0);
      glVertex2f(0,0);

      glTexCoord2f($image_w,0);
      glVertex2f($wnd_width,0);

      glTexCoord2f($image_w,$image_h);
      glVertex2f($wnd_width,$wnd_height);

      glTexCoord2f(0,$image_h);
      glVertex2f(0,$wnd_height);
    }
    glEnd ();
  }
  # Render tiles to FBO
  else
  {
    my $xa = ($x1+$x2)/2;
    my $ya = ($y1+$y2)/2;

    my $fx = $freqx;
    my $fy = $freqy;
    my $fmod = $modes->{$mode};
    if ($fmod > 1)
    {
      $fx *= $fmod;
      $fy *= $fmod;
    }

    my $tile_x = $fbo_w / $fx;
    my $tile_y = $fbo_h / $fy;

    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, $fb_ID);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    InitViewport($fbo_w,$fbo_h);

    for (my $i=0; $i<$fy; $i++)
    {
      for (my $j=0; $j<$fx; $j++)
      {
        my $x = $j*$tile_x;
        my $y = $i*$tile_y;

        glBegin (GL_TRIANGLE_STRIP);
        {
          glTexCoord2f($xa,$ya);
          glVertex2f($x,$y+$tile_y/2);

          glTexCoord2f($x0,$y0);
          glVertex2f($x,$y);

          glTexCoord2f($x1,$y1);
          glVertex2f($x+$tile_x/6,$y+$tile_y/2);

          glTexCoord2f($x2,$y2);
          glVertex2f($x+$tile_x/3,$y);

          glTexCoord2f($x0,$y0);
          glVertex2f($x+$tile_x/2,$y+$tile_y/2);

          glTexCoord2f($x1,$y1);
          glVertex2f($x+2*$tile_x/3,$y);

          glTexCoord2f($x2,$y2);
          glVertex2f($x+5*$tile_x/6,$y+$tile_y/2);

          glTexCoord2f($x0,$y0);
          glVertex2f($x+$tile_x,$y);

          glTexCoord2f($xa,$ya);
          glVertex2f($x+$tile_x,$y+$tile_y/2);
        }
        glEnd ();

        glBegin (GL_TRIANGLE_STRIP);
        {
          glTexCoord2f($xa,$ya);
          glVertex2f($x+$tile_x,$y+$tile_y/2);

          glTexCoord2f($x0,$y0);
          glVertex2f($x+$tile_x,$y+$tile_y);

          glTexCoord2f($x2,$y2);
          glVertex2f($x+5*$tile_x/6,$y+$tile_y/2);

          glTexCoord2f($x1,$y1);
          glVertex2f($x+2*$tile_x/3,$y+$tile_y);

          glTexCoord2f($x0,$y0);
          glVertex2f($x+$tile_x/2,$y+$tile_y/2);

          glTexCoord2f($x2,$y2);
          glVertex2f($x+$tile_x/3,$y+$tile_y);

          glTexCoord2f($x1,$y1);
          glVertex2f($x+$tile_x/6,$y+$tile_y/2);

          glTexCoord2f($x0,$y0);
          glVertex2f($x,$y+$tile_y);

          glTexCoord2f($xa,$ya);
          glVertex2f($x,$y+$tile_y/2);
        }
        glEnd ();
      }
    }

    # Tile FBO to window
    DisplayFBO();
  }
  glDisable($tex_mode);

  glutSwapBuffers();

  if ($fps && !$pause)
  {
    ($x0,$dx0) = nudge($x0,$dx0,$image_w);
    ($y0,$dy0) = nudge($y0,$dy0,$image_h);
    ($x1,$dx1) = nudge($x1,$dx1,$image_w);
    ($y1,$dy1) = nudge($y1,$dy1,$image_h);
    ($x2,$dx2) = nudge($x2,$dx2,$image_w);
    ($y2,$dy2) = nudge($y2,$dy2,$image_h);
  }
}

# Display FBO tiles
sub DisplayFBO
{
  glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
  glBindTexture($tex_mode, $fbo_ID);

  glEnable($tex_mode);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  my $tile_x = $wnd_width / $mfreqx;
  my $tile_y = $wnd_height / $mfreqy;

  for (my $i=0; $i<$mfreqy; $i++)
  {
    for (my $j=0; $j<$mfreqx; $j++)
    {
      my $wx0 = $j*$tile_x;
      my $wy0 = $i*$tile_y;
      my $wx1 = $wx0 + $tile_x;
      my $wy1 = $wy0 + $tile_y;

      # Draw FBO to window
      if (!$mode)
      {
        glBegin (GL_QUADS);
        {
          glTexCoord2f(0,0);
          glVertex2f($wx0,$wy0);

          glTexCoord2f($fbo_w,0);
          glVertex2f($wx1,$wy0);

          glTexCoord2f($fbo_w,$fbo_h);
          glVertex2f($wx1,$wy1);

          glTexCoord2f(0,$fbo_h);
          glVertex2f($wx0,$wy1);
        }
        glEnd ();
      }
      else
      {
        my $fx = $fbo_w / 2;
        my $fy = $fbo_h / 2;

        if ($mode == 1)
        {
          my $a = $ma * 2 * 3.1415;
          my $mx0 = $mx + $ml * cos($a);
          my $my0 = $my + $ml * sin($a);

          my $mx1 = $fx + $mx0;
          my $my1 = $fy + $my0;

          glBegin (GL_QUADS);
          {
            glTexCoord2f($mx0,$my0);
            glVertex2f($wx0,$wy0);

            glTexCoord2f($mx1,$my0);
            glVertex2f($wx1,$wy0);

            glTexCoord2f($mx1,$my1);
            glVertex2f($wx1,$wy1);

            glTexCoord2f($mx0,$my1);
            glVertex2f($wx0,$wy1);
          }
          glEnd ();

          if ($fps && !$pause)
          {
            $ma += $sgn * $mda;
            if ($ma > 1.0)
            {
              $ma -= 1.0;
            }
            elsif ($ma < 0.0)
            {
              $ma += 1.0;
            }
          }
        }
        elsif ($mode == 2)
        {
          my $wcx = $wnd_width / 2;
          my $wcy = $wnd_height / 2;

          my $mx0 = 0;
          my $my0 = 0;
          my $mx1 = $fbo_w;
          my $my1 = $fbo_h;
          my $mcx = $fbo_w / 2;
          my $mcy = $fbo_h / 2;

          glBegin (GL_TRIANGLES);
          {
            glTexCoord2f($mx0,$my0);
            glVertex2f($wx0,$wy0);

            glTexCoord2f($mcx,$mcy);
            glVertex2f($wx1,$wy0);

            glTexCoord2f($mx0,$my1);
            glVertex2f($wcx,$wcy);


            glTexCoord2f($mx1,$my1);
            glVertex2f($wx1,$wy0);

            glTexCoord2f($mcx,$mcy);
            glVertex2f($wx1,$wy1);

            glTexCoord2f($mx1,$my0);
            glVertex2f($wcx,$wcy);


            glTexCoord2f($mx0,$my0);
            glVertex2f($wx1,$wy1);

            glTexCoord2f($mcx,$mcy);
            glVertex2f($wx0,$wy1);

            glTexCoord2f($mx0,$my1);
            glVertex2f($wcx,$wcy);


            glTexCoord2f($mx1,$my1);
            glVertex2f($wx0,$wy1);

            glTexCoord2f($mcx,$mcy);
            glVertex2f($wx0,$wy0);

            glTexCoord2f($mx1,$my0);
            glVertex2f($wcx,$wcy);
          }
          glEnd ();
        }
      }
    }
  }

  glDisable($tex_mode);
}

# Clamp texcoords
sub nudge
{
  my($v,$d,$max) = @_;
  $v += $sgn * $d;
  return(abs($v),-$d) if ($v < 0);
  return(2*$max-$v,-$d) if ($v > $max);
  return($v,$d);
}

# Keyboard callback
sub cbKeyPressed
{
  my $key = shift;
  my $c = uc chr $key;
  if ($key == 27 or $c eq 'Q')
  {
    TermApp();
    exit(1);
  }
  elsif ($c eq 'P')
  {
    $pause = !$pause;
  }
  elsif ($c eq '*')
  {
    $fps = -1;
  }
  elsif (($c eq 'R') || ($c eq '+') || ($c eq '-'))
  {
    if ($c eq 'R')
    {
      $sgn *= -1.0;
    }
    elsif ($hasHires)
    {
      if ($c eq '+')
      {
        $fps = (($fps < 0) || ($fps >= 60)) ? -1 : $fps + 5;
      }
      elsif ($fps < 0)
      {
        $fps = 60;
      }
      elsif ($fps < 5)
      {
        $fps = 0;
      }
      else
      {
        $fps -= 5;
      }
      return;
    }
    else
    {
      if ($c eq '+')
      {
        $scl += 0.1;
      }
      elsif ($scl > 0.1)
      {
        $scl -= 0.1;
      }
      else
      {
        $scl = 0.0;
      }

      $dx0 = $scl * $ix0;
      $dy0 = $scl * $iy0;
      $dx1 = $scl * $ix1;
      $dy1 = $scl * $iy1;
      $dx2 = $scl * $ix2;
      $dy2 = $scl * $iy2;
    }
  }
  elsif ($c eq '0')
  {
    $equilateral=0;
    InitCoords();
  }
  elsif ($c eq 'E')
  {
    $equilateral=1;
    InitCoords();
  }
  elsif ($c eq 'O')
  {
    $display_source = !$display_source;
  }
  elsif ($c eq 'F')
  {
    ToggleFS();
  }
  elsif ($c eq 'M')
  {
    my $mcount = scalar(keys(%$modes));
    $mode = ($mode+1) % $mcount;
  }
  elsif ($c eq 'N')
  {
    LoadNextImage(1);
  }
  elsif ($c eq 'S')
  {
if (0)
{
    my $frame = new OpenGL::Image(engine=>'Magick',
      width=>$wnd_width, height=>$wnd_height);
    my($fmt,$size) = $frame->Get('gl_format','gl_type');
    glReadPixels_c( 0, 0, $wnd_width, $wnd_height, $fmt, $size, $frame->Ptr() );
    $frame->Save("$name.jpg");
}
else
{
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, $fb_ID);
    my $frame = new OpenGL::Image(engine=>'Magick',
      width=>$fbo_w, height=>$fbo_h);
    my($fmt,$size) = $frame->Get('gl_format','gl_type');
    glReadPixels_c( 0, 0, $fbo_w, $fbo_h, $fmt, $size, $frame->Ptr() );
    $frame->Save("$name.jpg");
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
}
  }
  elsif ($c ge '1' && $c le '9')
  {
    $freqx = $c;
    $freqy = $c;
  }
  elsif ($keys->{$c})
  {
    my $f = $keys->{$c};
    $mfreqx = $f;
    $mfreqy = $f;
  }
  else
  {
    printf "Key action undefined for %d.\n", $key;
  }
}

# Toggle fullscreen mode
sub ToggleFS
{
  my($enable) = @_;

  $fs = $enable ? 1 : !$fs;

  if ($fs)
  {
    $save_x = glutGet(GLUT_WINDOW_X);
    $save_y = glutGet(GLUT_WINDOW_Y);
    $save_w = $wnd_width;
    $save_h = $wnd_height;
    glutFullScreen();
    $wnd_width = glutGet(GLUT_SCREEN_WIDTH);
    $wnd_height = glutGet(GLUT_SCREEN_HEIGHT);
  }
  else
  {
    $wnd_width = $save_w;
    $wnd_height = $save_h;
    glutReshapeWindow($wnd_width,$wnd_height);
    glutPositionWindow($save_x,$save_y);
  }
}

# Window resize callback
sub cbResizeScene
{
  my ($Width,$Height) = @_;

  glViewport(0, 0, $Width, $Height);

  $wnd_width  = $Width;
  $wnd_height = $Height;
}

# Initialize Viewport
sub InitViewport
{
  my($w,$h) = @_;
  glViewport(0, 0, $w, $h);

  # Set up projection matrix
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluOrtho2D(0, $w, 0, $h);

  # Load identity modelview
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();

  # Shading states
  glShadeModel(GL_SMOOTH);
  glClearColor(0, 0, 0, 1);
  glColor4f(1.0, 1.0, 1.0, 1.0);
  glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);

  # Depth states
  glClearDepth(1.0);
  glDepthFunc(GL_LEQUAL);
  glEnable(GL_DEPTH_TEST);

  glEnable(GL_CULL_FACE);
}

# Cleanup routine
sub TermApp
{
  # Disable app
  glutHideWindow();
  glutKeyboardFunc();
  glutSpecialFunc();
  glutIdleFunc();
  glutReshapeFunc();

  # Release resources
  glBindRenderbufferEXT( GL_RENDERBUFFER_EXT, 0 );
  glBindFramebufferEXT( GL_FRAMEBUFFER_EXT, 0 );

  glDeleteRenderbuffersEXT_p( $rb_ID ) if ($rb_ID);
  glDeleteFramebuffersEXT_p( $fb_ID ) if ($fb_ID);

  glDeleteTextures_p($tex_ID,$fbo_ID);

  # Now you can destroy window
  glutDestroyWindow($wnd_ID);
}
